# ==================================================================
# Author: yvind Risan
# ==================================================================
# Input to "flood_delay.tcl": #Nodes #Active #Rounds DefaultWait opt(random)
#
# Tests the performance of the a pure flooding algorithm as described in yvind Risan's thesis of 2010
# Behaviour is to assign (random) a given number of nodes information to spread through the network.
# Each node represents a vehicle and will gather all information that is presented.
# The nodes move independently through the whole simulation.
# A simulation consist of a number of rounds, this number should be above 50 to produce stable results.
# The gathered information, delays, and other metrics are gathered and written to a result file.
# The input "DefaultWait" is not used, it is just kept to comply with the support script

# ======================================================================
# Define Simulation options
# ======================================================================
set optNodes 		50			;# defines the number of nodes
set optActive		10			;# number of active nodes
set optRounds 		10			;# defines the number of simulator rounds
set opt(interval)	10			;# Time between each round (seconds)
set DefaultWait		0.01		;# Default delay befor sending content
set opt(random)		1			;# Descide to make new random patterns
set optNam			0			;# Descide if NAM is to be used

# ======================================================================
# Define Environment options
# ======================================================================
set output_name		"output"	;# defines the name of the output file
set resultFile		"results/flood_Results.txt";# defines the name of the result file
set opt(sc)			"scene"		;# namesetting for the movment pattern file
set opt(speed)		10.0		;# defines the Max speed of the nodes (m/s)
set opt(pause)		0.10		;# defines the pause befor the nodes move
set opt(X)			1000		;# defines the X-size of the area
set opt(Y)			1000		;# defines the Y-size of the area
set opt(n-size)		20			;# defines the size of the displayed nodes

# ======================================================================
# Define Operation options
# ======================================================================
set opt(cbr_size) 		50000    				;# constant bit rate
set opt(cbr_interval) 	0.002					;# constant bit rate interval
set val(chan) 			Channel/WirelessChannel ;# channel type
set val(prop) 			Propagation/TwoRayGround;# radio-propagation model
set val(netif) 			Phy/WirelessPhy 		;# network interface type
set val(mac) 			Mac/802_11 				;# MAC type
set val(ifq) 			Queue/DropTail/PriQueue ;# interface queue type
set val(ll) 			LL 						;# link layer type
set val(ant) 			Antenna/OmniAntenna 	;# antenna model
set val(ifqlen) 		50 						;# max packet in ifq
set val(rp) 			DumbAgent 				;# routing protocol (DSDV, AODV, TORA or DSR)

# ======================================================================
# Define Command line input
# ======================================================================
if { $argc > 0 } {
	set optNodes 		[lindex $argv 0]
	set optActive		[lindex $argv 1]
	set optRounds 		[lindex $argv 2]
	set DefaultWait		[lindex $argv 3]
	set opt(random)		[lindex $argv 4]

	set output_name	"results/flood_trace_V$optNodes+_A$optActive+_D$DefaultWait"	;# defines the name of the output file
	set opt(sc) "sceneV$optNodes+_A$optActive"		;# defines the name of the movement file
}

set MESSAGE_PORT 42
set BROADCAST_ADDR -1
set opt(duration) [expr {$optRounds*$opt(interval)}] ;# duration of the simulation

array set RX 			{}
array set TX 			{}
array set FIRST 		{}
array set LAST 			{}
array set INFO 			{}
array set MESSAGE_SEEN	{}
array set RESULT 		{}
array set TIME 			{}
array set node_result 	{}

#
# Initialise ns_
#
set ns_ [new Simulator]
set tracefd [open $output_name.tr w]
$ns_ trace-all $tracefd

if { $optNam == 1 } {
	set nf [open $output_name.nam w]
	$ns_ namtrace-all-wireless $nf $opt(X) $opt(Y)
}

$ns_ use-newtrace

# ======================================================================
# prepare file for writing
# ======================================================================
#set fhandle [open ICE_Results w]

#puts $fhandle "---------------------------------------------------------"
#puts $fhandle "Number of rounds:		$optRounds"
#puts $fhandle "Number of nodes:		$optNodes"
#puts $fhandle "Number of active cells:		$optActive"
#puts $fhandle "Default delay before send:	[expr $DefaultWait *1000]\n"

#close $fhandle

# ======================================================================
# Define colour index
# ======================================================================
$ns_ color 0 blue
$ns_ color 1 red
$ns_ color 2 chocolate
$ns_ color 3 red
$ns_ color 4 brown
$ns_ color 5 tan
$ns_ color 6 gold
$ns_ color 7 black

#
# set up topography object
#
set topo       [new Topography]
$topo load_flatgrid $opt(X) $opt(Y)

#
#changes the queue type when running DSR, prevents "Segmentation fault (core dumped)"
#
if { $val(rp) == "DSR" } {
   set val(ifq)            CMUPriQueue
} else {
   set val(ifq)            Queue/DropTail/PriQueue
}

#
#creates the God entity
#
set god_ "God"				;# Bypases the problem with incorrect naming in the setdest file
create-god $optNodes

#generates random movment pattern
if { $opt(random) == "1"} {
	puts "*** NOTE: Random pattern generated."
        exec ./setdest.exe -n $optNodes -p $opt(pause) -M $opt(speed) -t $opt(duration) -x $opt(X) -y $opt(Y) > $opt(sc)
}

#
#configures the node template. Agent- and Mac-trace ON, Route- and Movement-trace OFF
#
$ns_ node-config -adhocRouting $val(rp) -llType $val(ll) \
     -macType $val(mac)  -ifqType $val(ifq) \
     -ifqLen $val(ifqlen) -antType $val(ant) \
     -propType $val(prop) -phyType $val(netif) \
     -channel  [new $val(chan)] -topoInstance $topo \
     -agentTrace OFF -routerTrace OFF\
     -macTrace ON -movementTrace OFF

#
#create the nodes
#
for {set i 0} {$i < $optNodes } {incr i} {
		set node_($i) [$ns_ node]
		$node_($i) color "black"

		set DISCARD($i) 0
		set RX($i) 0
		set TX($i) 0
		set FIRST($i) 0
		set LAST($i) 0
		set INFO($i) ""
		set MESSAGE_SEEN($i) ""
		set RESULT($i) ""
		set TIME($i) ""
		set node_result($i) 0
}

#
# Set the node-positions and speed at random, and initialise the nodes
#
if { $opt(sc) == "" } {
		puts "*** NOTE: no scenario file specified."
		set opt(sc) "none"
	} else {
		source $opt(sc)
		puts "Random positioning"
	}

for {set i 0} {$i < $optNodes } {incr i} {
	$ns_ initial_node_pos $node_($i) $opt(n-size)
}




# ======================================================================
# Subclass Agent/MessagePassing to make it do flooding
# ======================================================================
Class Agent/MessagePassing/Flooding -superclass Agent/MessagePassing

# ======================================================================
#Procedure for receiving messages
# ======================================================================
Class Agent/MessagePassing/Flooding -superclass Agent/MessagePassing

#
#Reciveing the messages
#
Agent/MessagePassing/Flooding instproc recv {source sport size data} {
	$self instvar messages_seen node_
	$self instvar Status node_

	global ns_ BROADCAST_ADDR RX TX DISCARD LAST FIRST INFO MESSAGE_SEEN optActive

	set Node_adr [$node_ node-addr]
	set now [$ns_ now]

	if {($FIRST($Node_adr) == 0)} {
		set ::FIRST($Node_adr) $now
	}

	set ::LAST($Node_adr) $now
	set ::RX($Node_adr) [expr $RX($Node_adr) + 1]

	#
	# extract Information from the message
	#
	set Timestamp 	[lindex [split $data ":"] 0]
	set Origin_ID 	[lindex [split $data ":"] 1]
	set Msg 		[lindex [split $data ":"] 2]

	#$ns_ trace-annotate "$Node_adr received {$data} from $source"
	#puts "$Node_adr received $data from $source"

	if {([lsearch $MESSAGE_SEEN($Node_adr) $Timestamp+$Origin_ID+$Msg] == -1)} {


		lappend ::MESSAGE_SEEN($Node_adr) $Timestamp+$Origin_ID+$Msg


	#
	# Detects new information and stores this
	#
	if {([lsearch $INFO($Node_adr) $Msg] == -1)} {
		lappend ::INFO($Node_adr) $Msg
		#$ns_ trace-annotate "$Node_adr received new {$Msg} from $source"
	}

	#
	# Makes the node green when the information is complete (visual aid only)
	#
	if {([llength $INFO($Node_adr)] == $optActive )} {

			$ns_ at $now "$node_ color green"
	}



		#$ns_ trace-annotate "$Node_adr received {$data} from $source"
		#puts "$Node_adr received $data from $source"


		$self sendto $size $data $BROADCAST_ADDR $sport
		set ::TX($Node_adr) [expr $TX($Node_adr) + 1]

	} else {
		#$ns_ trace-annotate "$Node_adr received redundant message $Timestamp+$Origin_ID+$ReplyID+$ReplyCond from $source"
		set ::DISCARD($Node_adr) [expr $DISCARD($Node_adr) + 1]

		#puts "$Node_adr received redundant message $Timestamp+$Origin_ID+$Msg from $source"
		#do nothing
	}
}

#
# ======================================================================
#Procedure for sending messages
# ======================================================================
Agent/MessagePassing/Flooding instproc send_message {size Timestamp Origin_ID Msg port} {
	$self instvar messages_seen node_
	global ns_ MESSAGE_PORT BROADCAST_ADDR TX MESSAGE_SEEN
	set Node_adr [$node_ node-addr]

	#$ns_ trace-annotate "[$node_ node-addr] sending message $Msg"
	#puts "[$node_ node-addr] sending message $Msg"
	$self sendto $size "$Timestamp:$Origin_ID:$Msg" $BROADCAST_ADDR $port
		incr ::TX($Node_adr)
	lappend ::MESSAGE_SEEN($Node_adr) $Timestamp+$Origin_ID+$Msg
}





#
# attach a new Agent/MessagePassing/Flooding to each node on port $MESSAGE_PORT
#
for {set i 0} {$i < $optNodes} {incr i} {
    set a($i) [new Agent/MessagePassing/Flooding]
    $node_($i) attach  $a($i) $MESSAGE_PORT
    $a($i) set messages_seen {}
    $a($i) set Trans_Time 0
    $a($i) set Delay $DefaultWait
}

#
#set up the events
#
set now [$ns_ now]

for {set j 0} {$j < $optRounds} {incr j} {
	for {set i 0} {$i < $optActive} {incr i} {
		set start_time [expr {(double(rand()*0.01)+($j*$opt(interval)))}]
		set random_node [expr {int(rand()*($optNodes-1))}]

		$ns_ at [expr {$j*$opt(interval)}] "lappend INFO($random_node) [expr $i+($j*100)]"

		$ns_ at $start_time "$a($random_node) send_message 10 $start_time $random_node [expr $i+($j*100)] $MESSAGE_PORT"
	}
	$ns_ at [expr (double($j*$opt(interval)+($opt(interval) / 2)))] "clear"
}

#
# Tell nodes when the simulation ends
#
for {set i 1} {$i < $optNodes } {incr i} {
	$ns_ at [expr $opt(duration) +10.0] "$node_($i) reset";
}

$ns_ at [expr $opt(duration) +10.0] "print"
$ns_ at [expr $opt(duration) +10.0] "finish"
$ns_ at [expr $opt(duration) +10.01] "puts \"ns_ Exiting...\"; $ns_ halt"

# ======================================================================
# Procedure to clear the information in each node
# ======================================================================
proc clear {} {

	global optNodes node_ INFO RESULT LAST FIRST TIME node_result

	for {set i 0} {$i < $optNodes} {incr i} {
			lappend RESULT($i) [llength $INFO($i)]
			set ::node_result($i) [expr ($node_result($i) + [llength $INFO($i)])]
			#puts "$i har  $INFO($i)"
			lset ::INFO($i) ""
			lset ::MESSAGE_SEEN($i) ""
			#lset ::messages_seen($i) ""
			lset ::TIME($i) [expr ($TIME($i) + ($LAST($i) - $FIRST($i))*1000)]
			lset ::FIRST($i) 0
			lset ::LAST($i) 0


			$node_($i) color blue
	}
}

# ======================================================================
# Procedure to print relevant information
# ======================================================================
proc print {} {
	global RX TX RESULT TIME optNodes optRounds node_result DefaultWait optActive resultFile

	set avr_sent 0.0
	set avr_time 0.0
	set avr_result 0.0

	for {set i 0} {$i < $optNodes} {incr i} {
		#puts " Vehicle$i	sent/received: $TX($i) / $RX($i) messages, was involved $TIME($i) ms"
		#puts " Set of received conditions:	$RESULT($i) \n"

		set avr_sent [expr ($avr_sent + $TX($i))]
		set avr_time [expr ($avr_time + $TIME($i))]
		set avr_result [expr ($avr_result + $node_result($i))]
	}

	puts " \n Number of Flood rounds:	$optRounds"
	puts " Number of nodes:		$optNodes"
	puts " Number of active cells:	$optActive"
	puts " Default delay before send:	[expr $DefaultWait *1000] ms"
	puts " Average sent pr node/round:	[expr $avr_sent / $optNodes / $optRounds]"
	puts " Average time pr node/round:	[expr $avr_time / $optNodes / $optRounds] ms"
	puts " Average result pr node/round:	[expr $avr_result/$optNodes/$optRounds] out of $optActive\n"

	set fhandle [open $resultFile a]
		puts $fhandle "Average sent pr node/round:		[expr $avr_sent / $optNodes / $optRounds]"
		puts $fhandle "Average time pr node/round:		[expr $avr_time / $optNodes / $optRounds]"
		puts $fhandle "Average result pr node/round:	[expr $avr_result/$optNodes/$optRounds]\n\n"
	close $fhandle
}

# ======================================================================
# Procedure to clean up and end the simulation
# ======================================================================
proc finish {} {
	global ns_ tracefd nf optNam output_name

	$ns_ flush-trace
	close $tracefd

	if { $optNam == 1 } {
		close $nf
		exec nam $output_name &
	}
exit 0
}
puts "Starting Simulation..."
$ns_ run